/*
 * <spd@daphne.cps.unizar.es>
 * Mon Apr 30 21:12:19 DST 2001
 * Sample tcp server using select()
 *
 * ex: set tabstop=4
 *
 */

static char __ident[] = "@(#)SPDsoft, Wed May  2 18:55:37 DST 2001";
#define VERS_STR        ((char*)&__ident[4])




#include <stdio.h>
#include <stdarg.h>  /* va_ */
#include <stdlib.h>  /* atoi */
#if defined(__MACH__) && defined(__APPLE__)
#include <stdint.h>
#endif
#include <unistd.h>  /* getpid */
#include <fcntl.h>   /* getpid */
#include <string.h>

#include <sys/time.h>  
#include <sys/types.h>  
#include <sys/socket.h> 
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>  

#include <sys/ioctl.h> 

#include <signal.h>
#include <sys/wait.h>

#include <errno.h>  
#include <assert.h>  

#include "network.h"

#define HAVE_SETSID 1 /* irix, linux, solaris */
#define HAVE_STRERROR 1

#ifndef  HAVE_STRERROR
#	define strerror(e) sys_errlist[(e)]
#endif

#ifndef _SYSV_SIGNALS
#	define _BSD_SIGNALS
#endif
#ifdef _BSD_SIGNALS
#	define sigset	signal
#endif

#define IPL	32
#define MAXS	100


/*
 * Globals
 */

int	done;

typedef	struct	{
	char	my_ip[IPL];
	int		port_d, port_s;
} server_data_struct;

server_data_struct	server_data[MAXS];

typedef	struct	{
	unsigned int		port;
	int		debug;
	int		udp;
	char	*bindh;
	char	*logfile;
	FILE	*f;
} prefs_struct;

prefs_struct	gprefs;

struct sockaddr_in		s_addr;
void (*sterm)(int);
void (*sint)(int);

/*
 * Prototypes
 */

void clean( int i );
void debug_msg(const char *fmt, ...);

/*
 */


#define perror(e) debug_msg("%s (%s)\n", e, strerror(errno))
void debug_msg(const char *fmt, ...)
{
	va_list                 vl;

	if ( NULL != gprefs.f)
	{
		va_start(vl,fmt);
			vfprintf(gprefs.f,fmt,vl);
		va_end(vl);
		fflush(gprefs.f);
	}
}


int start_listener(void)
{
	struct sockaddr_in		server;
	int						reuse=1;
	int						s=-1;

	gprefs.port=DEFAULT_TPORT;
	gprefs.debug=0;
	gprefs.logfile=NULL;
	gprefs.bindh=NULL;
	gprefs.f=NULL;


	memset(&server,0,sizeof(server));

	server.sin_family =			AF_INET;
	server.sin_port =			htons(gprefs.port);
	server.sin_addr.s_addr =	htonl(INADDR_ANY);

	if (gprefs.bindh != NULL )
		server.sin_addr.s_addr = inet_addr(gprefs.bindh);

	if (
		( SIG_ERR == (sterm=sigset(SIGTERM, clean))) ||
		( SIG_ERR == (sint=sigset(SIGINT, clean))) )
		perror("signal");


	if ( (s=socket(AF_INET,SOCK_DGRAM,0)) == -1 )
	{
		perror("Can't get socket");
		return(-1);
	}

	if (setsockopt(s, SOL_SOCKET,SO_REUSEADDR,&reuse,sizeof(int)) == -1)
	{
		perror("setsockopt");
	}

	debug_msg("bind to %s\n", inet_ntoa(server.sin_addr));

	if( bind( s, (struct sockaddr *)&server, sizeof(server)) != 0 )
	{	
		perror("bind failed");
		return(-1);
	}

		
	if ( NULL != gprefs.logfile )
	{
		if ( NULL == ( gprefs.f = fopen(gprefs.logfile, "wb")))
		{
			perror(optarg);
			gprefs.logfile = NULL;
		}
	}

	return s;
}

int stop_listener(int s)
{
	if ( -1 == shutdown(s,2))
		perror("shutdown");
	else
		debug_msg("fd %d shutdown\n",s);

	if ( -1 == close(s))
		perror("close");
	else
		debug_msg("fd %d closed\n",s);
	
	sigset(SIGTERM, sterm);
	sigset(SIGINT, sint);
	return(errno); /* Keep compiler happy */
}


int wait_for_answer(int s, unsigned int *data)
{
#define BS 1000

	struct timeval			tv;
	fd_set					mask,
							rmask;
	int						nfound;
	int						maxfd;

	struct sockaddr_in		client;
	int						clientlen=sizeof(client);

	long					nbytes;
	unsigned char			buf[BS];

	unsigned int			now,start;
	int						late=0;

	gprefs.port=DEFAULT_TPORT;
	gprefs.debug=0;
	gprefs.logfile=NULL;
	gprefs.bindh=NULL;
	gprefs.f=stderr;


	tv.tv_sec = 60;
	tv.tv_usec = 500000;

	done=0;
	FD_ZERO(&mask);
	FD_SET(s, &mask);
	maxfd=s;


	for(;!done;)
	{
		rmask=mask;

/*		debug_msg("select in call s=%d\n",s);*/
		nfound=select(maxfd+1, &rmask, NULL, NULL, &tv);
#if defined __linux__
		/* Linux modifies tv, so we need to reset it */
		tv.tv_sec = 60;
		tv.tv_usec = 500000;
#endif

		if (nfound<0)
		{
			if ( EINTR == errno )
			{
				debug_msg("Interrupted system call\n");
				done=1;
				break;
			}
			/*
			 * An error occurred during the select, or
			 * the select timed out.
			 */
			if ( 0 != errno )
			{
				perror("select");
				done=1;
			}
		}
		if (nfound==0)
		{
			continue;
		}
		if (FD_ISSET(s, &rmask))
		{
			done=0;


			clientlen=sizeof(client);

			debug_msg("%d receiving\n", getpid());
			nbytes = recvfrom(
				s,
				buf,sizeof(buf)-1,0,
				(struct sockaddr *)&client,&clientlen);
			debug_msg("%d received %d bytes (0x%X) from %s via fd %d\n",
				getpid(), nbytes, *(unsigned*)buf,
				inet_ntoa(client.sin_addr), s);

			start = ntohl(*(unsigned *)buf);
			now=GetTimerMS();

			debug_msg("%d: start (%u) from %s. Now %u\n",
				getpid(), start, inet_ntoa(client.sin_addr), now);

			if ( (now+500) >= start )
			{
				done=1;
				sendto(
				s, "LATE  ", 6, 0, (struct sockaddr *) &client,
				sizeof(struct sockaddr_in));
				late=1;
			}
			else
			{
				done=1;
				sendto(
				s, "OK    ", 6, 0, (struct sockaddr *) &client,
				sizeof(struct sockaddr_in));
				late=0;
			}

			FD_ZERO(&mask);
			FD_SET(s, &mask);
		}
	}

	*data = start;
	return(late);
}

/************************************************************************/

void clean( int sig )
{
	if ( sig != SIGSEGV)
	{
		if ( SIG_ERR == sigset(sig, clean))
			perror("clean:signal");

		debug_msg("%d killed by signal (%d)\n", getpid(), sig);
		fflush(stderr);
		done=1;
	}
}


/************************************************************************/

int open_udp_server(
	char *host,
	unsigned int port,
	unsigned timeout
	)
{
	int				sockfd;
	int				i=1;
	unsigned long	addr;
	struct hostent	*he;
	struct sockaddr_in	peer;
	struct sockaddr_in	local;

	alarm(timeout);

	/*
	 * Broadcast: we need a datagram UDP socket
	 */
	if ( 0 > (sockfd = socket(AF_INET,SOCK_DGRAM,0)))
	{
		perror("client: can't open stream socket");
		alarm(0);
		return(-1);
	}

	addr = inet_addr(host);
	if (addr == INADDR_NONE)
	{
		he = gethostbyname(host);
		if (he == NULL)
		{
			perror("gethostbyname: no IP address found");
			alarm(0);
			return(-1);
		}
		addr = ((struct in_addr *)(he->h_addr))->s_addr;
	}

	memset(&peer,0,sizeof(peer));
	peer.sin_family       = AF_INET;
	peer.sin_addr.s_addr  = addr;
	peer.sin_port         = htons(port);

	memset(&local,0,sizeof(local));
	local.sin_family       = AF_INET;
	local.sin_addr.s_addr  = INADDR_ANY;
	local.sin_port         = 0;

	if (bind(sockfd, (struct sockaddr *) &local, sizeof(local)))
	{
		perror("client: can't bind socket");
		alarm(0);
		close(sockfd);
		return(-1);
	}

	if ( 0 > ( setsockopt(
			sockfd, SOL_SOCKET,SO_BROADCAST, (void *)&i,sizeof(i))))
	{
		perror("client: failure to set SO_BROADCAST");
		close(sockfd);
		alarm(0);
		return(-1);
	}

	memcpy(&s_addr,&peer,sizeof(peer));
	alarm(0);
	return sockfd;
}

int open_tcp_server(
	char *host,
	unsigned int port,
	unsigned timeout
	)
{
	int				sockfd;
	int				i=1;
	unsigned long	addr;
	struct hostent	*he;
	struct sockaddr_in	peer;
	struct sockaddr_in	local;

	alarm(timeout);

	/*
	 * Broadcast: we need a datagram UDP socket
	 */
	if ( 0 > (sockfd = socket(AF_INET,SOCK_STREAM,0)))
	{
		perror("client: can't open stream socket");
		alarm(0);
		return(-1);
	}

	addr = inet_addr(host);
	if (addr == INADDR_NONE)
	{
		he = gethostbyname(host);
		if (he == NULL)
		{
			perror("gethostbyname: no IP address found");
			alarm(0);
			return(-1);
		}
		addr = ((struct in_addr *)(he->h_addr))->s_addr;
	}

	memset(&peer,0,sizeof(peer));
	peer.sin_family       = AF_INET;
	peer.sin_addr.s_addr  = addr;
	peer.sin_port         = htons(port);

	memset(&local,0,sizeof(local));
	local.sin_family       = AF_INET;
	local.sin_addr.s_addr  = INADDR_ANY;
	local.sin_port         = 0;

	if (bind(sockfd, (struct sockaddr *) &local, sizeof(local)))
	{
		perror("client: can't bind socket");
		alarm(0);
		close(sockfd);
		return(-1);
	}

	if ( 0 > connect(sockfd,(struct sockaddr *)&peer, sizeof(peer)))
	{
			perror("client: can't connect to server");
			close(sockfd);
			alarm(0);
			return(-1);
	}

	memcpy(&s_addr,&peer,sizeof(peer));
	alarm(0);
	return sockfd;
}
/*****************************************************************************/
int close_tcp_server(int fd)
{
	return(close(fd));
}
/*****************************************************************************/
int send_bcast(
	int                 sockfd,
	void                *data,
	int                 size)
{
	int				n;

	if ( 0 >= ( n=sendto(
		sockfd, data, size, 0, (struct sockaddr *) &s_addr,
		sizeof(struct sockaddr_in))))
	{
		perror("client: sendto failed");
		return -1;
	}
	else
	{
		return 0;
	}
}

int recv_bcast(int s, int answers, unsigned timeout)
{
	long					nbytes;
	struct sockaddr_in		server,
							client;
	int						clientlen=sizeof(client);
	unsigned char			buf[BS];

	int						done;

	struct timeval			tv;
	fd_set					mask,
							rmask;
	int						nfound;
	int						maxfd;


	tv.tv_sec = 2;
	tv.tv_usec = 500000;

	done=0;
	FD_ZERO(&mask);
	FD_SET(s, &mask);
	maxfd=s;

	for(;done<answers;)
	{
		rmask=mask;

		nfound=select(maxfd+1, &rmask, NULL, NULL, &tv);

		if (nfound<0)
		{
			if ( EINTR == errno )
			{
				debug_msg("Interrupted system call\n");
				done=1;
				break;
			}
			/*
			 * An error occurred during the select, or
			 * the select timed out.
			 */
			if ( 0 != errno )
			{
				perror("select");
				done=1;
			}
		}
		if (nfound==0)
		{
			continue;
		}
		if (FD_ISSET(s, &rmask))
		{

			debug_msg("%d receiving\n", getpid());
			nbytes = recvfrom(
				s,
				buf,sizeof(buf)-1,0,
				(struct sockaddr *)&client,&clientlen);
			debug_msg("%d received %d bytes (%s) from %s\n",
				getpid(), nbytes, buf,
				inet_ntoa(client.sin_addr));

#if 0
			if ( 0 >= ( nbytes = sendto(
				s, "ACK   ", 6, 0, (struct sockaddr *) &s_addr,
				sizeof(struct sockaddr_in))))
			{
				perror("client: sendto failed");
			}
#endif
			done++;

			FD_ZERO(&mask);
			FD_SET(s, &mask);
		}
	}

	return(errno); /* Keep compiler happy */
}
